{-# LANGUAGE RecordWildCards #-}

import Affection
import qualified SDL
import qualified GEGL as G
import qualified BABL as B
import qualified Data.Map.Strict as M

import Foreign.C.Types

import System.Random (randomRIO)

import Debug.Trace

main :: IO ()
main = do
  conf <- return $ AffectionConfig
    { initComponents = All
    , windowTitle    = "Affection: example00"
    , windowConfig   = SDL.defaultWindow
    , preLoop        = drawInit
    , drawLoop       = draw
    , updateLoop     = update
    , loadState      = load
    , cleanUp        = clean
    }
  withAffection conf

data UserData = UserData
  { nodeGraph   :: M.Map String G.GeglNode
  , foreground  :: G.GeglBuffer
  , coordinates :: Maybe (Int, Int)
  , partsys     :: ParticleSystem
  }

load :: SDL.Surface -> IO UserData
load _ = do
  traceM "loading"
  root <- G.gegl_node_new
  traceM "new root node"
  checkerboard <- G.gegl_node_new_child root $ G.checkerboardOperation
    [ G.Property "color1" $ G.PropertyColor $ G.RGBA 0.4 0.4 0.4 1
    , G.Property "color2" $ G.PropertyColor $ G.RGBA 0.6 0.6 0.6 1
    ]
  traceM "checkerboard"
  over <- G.gegl_node_new_child root G.defaultOverOperation
  traceM "over"
  buffer <- G.gegl_buffer_new (G.GeglRectangle 0 0 800 600) =<<
    B.babl_format (B.PixelFormat B.RGBA B.CFfloat)
  bufsrc <- G.gegl_node_new_child root $ G.bufferSourceOperation
    [ G.Property "buffer" $ G.PropertyBuffer buffer
    ]
  traceM "buffer-source"
  G.gegl_node_link checkerboard over
  G.gegl_node_connect_to bufsrc "output" over "aux"
  traceM "connections made"
  myMap <- return $ M.fromList
    [ ("root"        , root)
    , ("over"        , over)
    , ("background"  , checkerboard)
    , ("foreground"  , bufsrc)
    ]
  traceM "loading complete"
  return $ UserData
    { nodeGraph = myMap
    , foreground = buffer
    , coordinates = Nothing
    , partsys = ParticleSystem [] over buffer
    }

drawInit :: Affection UserData ()
drawInit = do
  UserData{..} <- getAffection
  present (nodeGraph M.! "over") foreground (GeglRectangle 0 0 800 600) True

draw :: Affection UserData ()
draw = do
  traceM "drawing"
  -- ad <- get
  -- ud <- getAffection
  -- drawParticles partDraw $ particles ud
  -- SDL.V2 (CInt rw) (CInt rh) <- SDL.surfaceDimensions $ drawSurface ad
  -- let (w, h) = (fromIntegral rw, fromIntegral rh)
  -- liftIO $ clearArea (foreground ud) (GeglRectangle 0 0 w h)
  -- maybe (return ()) (\(x, y) ->
  --   drawRect
  --     (foreground ud)
  --     (nodeGraph ud M.! "over")
  --     (G.RGBA 1 0 0 0.5)
  --     (Line 7)
  --     (G.GeglRectangle (x - 10) (y - 10) 20 20)
  --   ) $ coordinates ud

update :: Double -> AffectionState (AffectionData UserData) IO ()
update sec = do
  traceM "updating"
  ad <- get
  ud <- getAffection
  -- let newPart = updateParticles sec partUpd $ particles ud
  -- putAffection $ ud { particles = newPart }
  nps <- updateParticleSystem (partsys ud) sec partUpd partDraw
  putAffection $ ud { partsys = nps }
  traceM $ (show $ 1 / sec) ++ " FPS"
  ev <- liftIO $ SDL.pollEvents
  mapM_ (\e ->
    case SDL.eventPayload e of
      SDL.MouseButtonEvent dat -> do
          let (SDL.P (SDL.V2 x y)) = SDL.mouseButtonEventPos dat
          vx <- liftIO $ randomRIO (-20, 20)
          vy <- liftIO $ randomRIO (-20, 20)
          life <- liftIO $ randomRIO (1, 5)
          traceM $ "velocity is: " ++ show vx ++ " " ++ show vy
          putAffection $ ud
            { coordinates = Just (fromIntegral x, fromIntegral y)
            , partsys     = (partsys ud)
                { psParts = (Particle
                    { particleLife = life
                    , particlePosition = (fromIntegral x, fromIntegral y)
                    , particleRotation = 0
                    , particleVelocity = (vx, vy)
                    }) : (psParts $ partsys ud)
                }
            }
      SDL.WindowClosedEvent _ -> do
        traceM "seeya!"
        put $ ad
          { quitEvent = True
          }
      _ ->
        return ()
    ) ev

clean :: UserData -> IO ()
clean _ = return ()

partUpd :: Double -> Particle -> Particle
partUpd sec p@Particle{..} =
  p
    { particlePosition = (newX, newY)
    , particleLife = particleLife - sec
    }
  where
    newX = (fst particlePosition) + sec * (fromIntegral $ fst particleVelocity)
    newY = (snd particlePosition) + sec * (fromIntegral $ snd particleVelocity)

partDraw :: G.GeglBuffer -> G.GeglNode -> Particle -> Affection UserData ()
partDraw buf node Particle{..} = do
  ud <- getAffection
  drawRect
    buf
    node
    (G.RGBA 1 0 0 0.5)
    (Line 5)
    (G.GeglRectangle ((floor $ fst particlePosition) - 10) ((floor $ snd particlePosition) -10) 20 20)
